#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>


enum makefirtype
{
	LP = 0,
	BP = 1,
	HP = 2,
	BR = 3
};

enum windowtype
{
	HANN = 0,
	HAMMING = 1,
	BLACKMAN = 2
};

template <typename DT = double>
class makefir
{
	makefirtype type;

	int order;
	DT forder;
	DT forder2;
	bool odd;

	DT lower, upper, bw;
	windowtype wt;
	DT beta;

	DT* coefficients;

	static constexpr double twopi = 2.0 * M_PI;

	static DT sinc(DT x)
	{
		if((x > DT(-1.0e-6)) && (x < DT(1.0e-6)))
			return(DT(1.0));
		return(sin(x)/x);
	}

	void NOINLINE
	dowindow(void)
	{
		const int orderp1 = order + 1;
		DT tmpcoeff[orderp1];
		DT orderinv = DT(twopi) / DT(orderp1);

		switch(wt)
		{

case HANN:
   		for(int idx = 0; idx < orderp1; idx++)
			{
				DT tmpval = DT(0.5) * (DT(1.0) - cos(DT(idx) * orderinv));
				tmpcoeff[idx] = tmpval;
			}
			break;

case HAMMING:
   		for(int idx = 0; idx < orderp1; idx++)
			{
				DT tmpval = DT(0.538)  - DT(0.462) * cos(DT(idx) * orderinv);
				tmpcoeff[idx] = tmpval;
			}
			break;

case BLACKMAN:
   		for(int idx = 0; idx < orderp1; idx++)
			{
				DT tmpval = DT(0.42)  - DT(0.5) * cos(DT(idx) * orderinv) + DT(0.08) * cos(DT(2.0) * DT(idx) * orderinv);
				tmpcoeff[idx] = tmpval;
			}

			break;
		}

#if 0
		DT gainsum{0.0};
   	for(int idx = 0; idx < orderp1; idx++)
		{
			gainsum += tmpcoeff[idx];
		}
		gainsum /= DT(orderp1);

		DT gainsuminv = (gainsum != DT(0.0)) ? (DT(1.0) / gainsum) : DT(1.0);
   	for(int idx = 0; idx < orderp1; idx++)
		{
			tmpcoeff[idx] *= gainsuminv;
		}
#endif

		for (int idx = 0; idx < orderp1; idx++)
		{
			coefficients[idx] *= tmpcoeff[idx];
		}

	}

public:
	makefir(makefirtype typearg, int orderarg, DT lowerarg, DT upperarg, windowtype wtarg = HANN, DT betaarg = DT(0.0)) NOINLINE :
		type(typearg), order(orderarg), lower(lowerarg), upper(upperarg), bw(upperarg - lowerarg),
		wt(wtarg), beta(betaarg), coefficients(nullptr)
	{
		order = orderarg;
		odd = (order & 1);
		forder = order;
		forder2 = DT(order - 1) / DT(2.0);
		coefficients = new DT[order + 1 + sizeof(double)];

		for (int i = 0; i < order + 1; i++)
			coefficients[i] = DT(0.0);

		switch(type)
		{
	case LP:
			{
				//fprintf(stderr,"LP: lowerarg: %lf\n", double(lowerarg));
				for (int idx = 0; idx < order + 1; idx++)
				{
					if (idx == order / 2)
					{
						coefficients[idx] = DT(2.0) * lower;
					}
					else
					{
						DT sarg = DT(idx) - forder2;
						DT sargpi = sarg * DT(M_PI);
						DT sarg2pi = sargpi * DT(2.0);
						coefficients[idx] = sin( sarg2pi * lower) / sargpi;
					}
				}
				break;
			}
	case HP:
			{
				for (int idx = 0; idx < order + 1; idx++)
				{
					if (idx == order / 2)
					{
						coefficients[idx] = (DT(1.0) - DT(2.0) * lower);
					}
					else
					{
						DT sarg = DT(idx) - forder2;
						DT sargpi = sarg * DT(M_PI);
						DT sarg2pi = sargpi * DT(2.0);
						coefficients[idx] = DT(-1.0) * sin( sarg2pi * lower) / sargpi;
					}
					//coefficients[idx] *= DT(-1.0);
				}
				break;
			}
				
	case BP:
				//fprintf(stderr,"BP: lowerarg: %lf, upperarg: %lf\n", double(lowerarg), double(upperarg));
				for (int idx = 0; idx < order + 1; idx++)
				{
					if (idx == order / 2)
					{
						coefficients[idx] = DT(2.0) * (upper - lower);
					}
					else
					{
						DT sarg = DT(idx) - forder2;
						DT sargpi = sarg * DT(M_PI);
						DT sarg2pi = DT(2.0) * sargpi;
						coefficients[idx] = DT(1.0) * (sin(sarg2pi * upper) - sin( sarg2pi * lower)) / sargpi;
					}
				}
				break;

	case BR:
				for (int idx = 0; idx <= order; idx++)
				{
					if (idx == order / 2)
					{
						coefficients[idx] = DT(1.0) - DT(2.0) * (upper - lower);
					}
					else
					{
						DT sarg = DT(idx) - forder2;
						DT sargpi = sarg * DT(M_PI);
						DT sarg2pi = DT(2.0) * sargpi;
						coefficients[idx] = (sin( sarg2pi * lower) - sin( sarg2pi * upper)) / sargpi;
					}
				}
				break;
			}

#if 1
		dowindow();
#else
		for (int idx = 0; idx <= order; idx++)
			coefficients[idx] *= DT(2.0);
#endif

	}

	~makefir(void)
	{
		delete[] coefficients;
	}

	void
	getcoeffs(DT *carray) const
	{
		for (int i = 0; i <= order; i++)
			carray[i] = coefficients[i];
	}

	const DT*
	coeffs(void) const
	{
		return coefficients;
	}
	
};

	
#if 0
template <typename DT = double>
class makefirBR
{
	makefirtype type;

	int order;
	DT forder;
	DT forder2;
	bool odd;

	DT lower, upper, upperbr;

	DT* coefficients;

	static constexpr double twopi = 2.0 * M_PI;

	static DT sinc(DT x)
	{
		if((x > DT(-1.0e-6)) && (x < DT(1.0e-6)))
			return(DT(1.0));
		return(sin(x)/x);
	}

	void NOINLINE
	dowindow(void)
	{
		DT *tmpcoeff = new DT[order + 1];
		DT orderinv = DT(twopi) / DT(order + 1);

   	for(int idx = 0; idx < (order / 2) + 1; idx++)
		{
			DT tmpval = DT(0.5) * (DT(1.0) - cos(DT(idx) * orderinv));
			tmpcoeff[idx] = tmpval;
		}
		for (int idx = (order / 2) + 1 - 1; idx <= order; idx++)
			tmpcoeff[idx] = tmpcoeff[order - idx - 1];

//			tmpcoeff[order - idx - 1] = tmpval;

#if 1
		DT gainsum{0.0};
   	for(int idx = 0; idx <= order; idx++)
		{
			gainsum += tmpcoeff[idx];
		}
		gainsum /= DT(order + 1);

		DT gainsuminv = (gainsum != DT(0.0)) ? (DT(1.0) / gainsum) : DT(1.0);
   	for(int idx = 0; idx <= order; idx++)
		{
			tmpcoeff[idx] *= gainsuminv;
		}
#endif

		for (int idx = 0; idx <= order; idx++)
		{
			coefficients[idx] *= tmpcoeff[idx];
		}

		delete[] tmpcoeff;
	}
		

public:
	makefirBR(int orderarg, DT lowerarg, DT upperarg, DT upperbrarg, DT nr = DT(1.0)) :
		order(orderarg), lower(lowerarg), upper(upperarg), upperbr(upperbrarg), coefficients(nullptr)
	{
		order = orderarg;
		odd = (order & 1);
		forder = order;
		forder2 = DT(order - 1) / DT(2.0);
		coefficients = new DT[order + 1 + sizeof(double)];

			for (int idx = 0; idx <= order; idx++)
			{
					DT sarg = DT(idx) - forder2;
					DT sargpi = sarg * DT(M_PI);
					DT sarg2pi = sargpi * DT(2.0);
				//coefficients[idx] = DT(0.50) * sinc(sargpi) - (upper * sinc(upper * sarg2pi) - lower * sinc(lower * sarg2pi)) - upperbr * sinc(upperbr * sarg2pi);
				//coefficients[idx] = upperbr * sinc(upperbr * sarg2pi) - upper * sinc(upper * sarg2pi) - lower * sinc(lower * sarg2pi);
				coefficients[idx] = DT(0.50) * sinc(sargpi) - upper * sinc(upper * sarg2pi) - lower * sinc(lower * sarg2pi);
				coefficients[idx] += nr * ((upper * sinc(upper * sarg2pi) - lower * sinc(lower * sarg2pi)));
			}

#if 1
		dowindow();
#else
		for (int idx = 0; idx <= order; idx++)
			coefficients[idx] *= DT(2.0);
#endif

	}

	~makefirBR(void)
	{
		delete[] coefficients;
	}

	void
	getcoeffs(DT *carray) const
	{
		for (int i = 0; i <= order; i++)
			carray[i] = coefficients[i];
	}

	const DT*
	coeffs(void) const
	{
		return coefficients;
	}
	
};
#endif

	
	
//
// vi: set ts=2:
//
